-- dcf77.vhd	ajm						25-nov-2002
--
-- entity	dcf77	-DCF77 sender
--			-(optional) preset signals and noise
-- architecture	simModel
--
-- note		Simulating noisy input, the uniform function from
--		ieee.math_real is used.			-> ncsim, leapfrog
--		This package is not (fully) implemented for Synopsys!
--		Use synopsys.distributions instead	-> scirocco, vss	
------------------------------------------------------------------------------
-- library synopsys;		-- Synopsys
-- use synopsys.distributions.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.math_real.all;		-- IEEE 1076.2

-- WARNING: tSample must be natural divider of 100 ms!
entity dcf77 is
generic (	tDelay	: time		:= 300 us;	-- initial Delay
		tSample	: time		:=   1 ms);	-- 1 kHz sample freq.
port	(	dcfsig	: out std_logic;		-- DCF77 signal
		mins	: in  integer range 0 to 59	:= 34;
		hours	: in  integer range 0 to 23	:= 12;
		wday	: in  integer range 1 to  7	:= 1;
		day	: in  integer range 1 to 31	:= 25;
		month	: in  integer range 1 to 12	:= 11;
		year	: in  integer range 0 to 99	:= 2;
		noise	: in  integer range 0 to 50	:= 0;
		xtra	: in  std_logic_vector (15 to 19):= "00000");
end entity dcf77;

------------------------------------------------------------------------------
architecture simModel of dcf77 is
  constant samp100ms	: integer	:= 100 ms/tSample; -- ticks per 100 ms


  signal dcfString		: std_logic_vector(0 to 58);
begin
  stringP: process is
    variable dcfVar		: std_logic_vector(0 to 58);
    variable minCnt, hourCnt	: integer;
    variable minsL,  hoursL	: integer	:= -1;	-- last values

    function  int2dcf (	arg	: natural;
			size	: positive)
			return	  std_logic_vector is
      variable	retVal		: std_logic_vector (1 to size);
      variable	tmp		: natural := arg;
    begin
      for i in retVal'range loop
	if (tmp mod 2) = 0	then	retVal(i) := '0';
				else	retVal(i) := '1';
	end if;
	tmp := tmp / 2;
      end loop;
      return retVal;
    end function int2dcf;

    function  parity (	vec	: std_logic_vector )
			return	  std_logic is
      variable	retVal		: std_logic := '0';
    begin
      for i in vec'range loop
	retVal := retVal xor vec(i);
      end loop;
      return retVal;
    end function parity;

  begin
    -- loop forever			--------------------------------------
    --------------------------------------------------------------------------
    loop
      if (mins /= minsL) or (hours /= hoursL) then	-- input changes?
	minCnt	:= mins;
	minsL	:= mins;
	hourCnt	:= hours;
	hoursL	:= hours;
      elsif minCnt = 59					-- count + 1 min
      then	minCnt := 0;
		if hourCnt = 23	then	hourCnt := 0;
				else	hourCnt := hourCnt + 1;
		end if;
      else	minCnt := minCnt + 1;
      end if;

      -- dcfString: formatting		--------------------------------------
      -- use sequential variable, otherwise parity will fail!
      ------------------------------------------------------------------------
      dcfVar( 0 to 14)	:= (others=>'0');
      dcfVar(15 to 19)	:= xtra;
      dcfVar(20)	:= '1';
      dcfVar(21 to 24)	:= int2dcf(minCnt rem 10, 4);
      dcfVar(25 to 27)	:= int2dcf(minCnt  /  10, 3);
      dcfVar(28)	:= parity(dcfVar(21 to 27));
      dcfVar(29 to 32)	:= int2dcf(hourCnt rem 10, 4);
      dcfVar(33 to 34)	:= int2dcf(hourCnt  /  10, 2);
      dcfVar(35)	:= parity(dcfVar(29 to 34));
      dcfVar(36 to 39)	:= int2dcf(day rem 10, 4);
      dcfVar(40 to 41)	:= int2dcf(day  /  10, 2);
      dcfVar(42 to 44)	:= int2dcf(wday, 3);
      dcfVar(45 to 48)	:= int2dcf(month rem 10, 4);
      if (month / 10) = 1	then dcfVar(49) := '1';
				else dcfVar(49) := '0';
      end if;
      dcfVar(50 to 53)	:= int2dcf(year rem 10, 4);
      dcfVar(54 to 57)	:= int2dcf(year  /  10, 4);
      dcfVar(58)	:= parity(dcfVar(36 to 57));
      dcfString		<= dcfVar;
      wait for 60 sec;
    end loop;				-- forever
  end process stringP;

  dcfP: process is
--  variable synSeed		: real				:= 1.82736459;
    variable seed1		: positive			:= 123456789;
    variable seed2		: positive			:= 8273645;
    variable random		: real;

    procedure sendNoisy(target	: std_logic;
			msCnt	: integer range 1 to 10) is
    begin
      for s in 1 to msCnt*samp100ms loop
--	uniform(synSeed, 0.0, 100.0, random);
--	if random < real(noise)
	uniform(seed1, seed2, random);
	if random*100.0 < real(noise)
	  then	dcfsig <= not target;	-- dcfsig: noisy
	  else	dcfsig <= target;	--  -"-  : OK
	end if;
	wait for tSample;
      end loop;
    end procedure sendNoisy;

  begin
    dcfsig  <= '1';
    wait for tDelay;

    -- loop forever			--------------------------------------
    --------------------------------------------------------------------------
    loop
      for i in dcfString'range loop	-- output computed DCF string
	if noise = 0 then		-- zero noise
	  if dcfString(i) = '1'	then	-- dcf '1'
	    dcfsig <= '0';
	    wait for 200 ms;
	    dcfsig <= '1';
	    wait for 800 ms;
	  else				-- dcf '0'
	    dcfsig <= '0';
	    wait for 100 ms;
	    dcfsig <= '1';
	    wait for 900 ms;
	  end if;			-- zero noise
        else				-- noise% errors at '0'/'1' level
	  if dcfString(i) = '1'	then	-- noisy '1'
	    sendNoisy('0', 2);
	    sendNoisy('1', 8);
	  else				-- noisy '0'
	    sendNoisy('0', 1);
	    sendNoisy('1', 9);
	  end if;
	end if;				-- noise% errors
      end loop;				-- DCF string
      -- now the "missing" second	--------------------------------------
      if noise = 0 then			-- zero noise
	wait for 1 sec;
      else				-- noise% errors at '0'/'1' level
	sendNoisy('1', 10);
      end if;				-- random noise
    end loop;				-- forever
  end process dcfP;
end architecture simModel;

-- sample usage		------------------------------------------------------
------------------------------------------------------------------------------
--library ieee;
--use ieee.std_logic_1164.all;
--
--entity xxx is
--end entity xxx;
--architecture tst of xxx is
--component dcf77 is
--generic (	tDelay	: time		:= 300 us;	-- initial Delay
--		tSample	: time		:=   1 ms);	-- 1 kHz sample freq.
--port	(	dcfsig	: out std_logic;		-- DCF77 signal
--		mins	: in  integer range 0 to 59	:= 34;
--		hours	: in  integer range 0 to 23	:= 12;
--		wday	: in  integer range 1 to  7	:= 1;
--		day	: in  integer range 1 to 31	:= 25;
--		month	: in  integer range 1 to 12	:= 11;
--		year	: in  integer range 0 to 99	:= 2;
--		noise	: in  integer range 0 to 50	:= 0;
--		xtra	: in  std_logic_vector (15 to 19):= "00000");
--end component dcf77;
--  signal xxx : std_logic;
--begin
--  sendI: dcf77 port map (xxx);
--end architecture tst;
